unit xutils;

{$mode delphi}

// xutils - X Window System utilities
//
// This is a work in progress, and more routines will be added when I have
// the time. Most of these routines are wrappers for Xlib routines.
// The idea is to make the C functions easier to call and use in Kylix programs.
//
// Copyright  2001 Tempest Software, Inc.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

// FREEPASCAL VERSION 2008

interface

uses Sysutils, x, xlib, xatom, Classes;

type
  // Callback procedure for EnumTopLevelWindows.
  // Dpy is the X-Windows display.
  // Win is the top-level window.
  // Param is the caller-supplied parameter.
  TWindowProc = procedure(Dpy: PDisplay; Win: TWindow; Param: Pointer) of object;
  // for ChangeProperty
  TPropMode = (pmReplace, pmPrepend, pmAppend);
  TWindowArray = array[0..MaxListSize] of TWindow;
  PWindowArray = ^TWindowArray;
  TAtomArray = array[0..MaxListSize] of TAtom;
  PAtomArray = ^TAtomArray;


// Change a property value.
function ChangeProperty(Dpy: PDisplay; Win: TWindow; Prop: TAtom; Value: Cardinal; Mode: TPropMode=pmReplace): Boolean; overload;

// Enumerate all top-level windows.
procedure EnumTopLevelWindows(Proc: TWindowProc; Param: Pointer);

// XGetWindowProperty is general, but clumsy. The following functions are
// easier to use for many common situations.
function GetProperty(Dpy: PDisplay; Win: TWindow; Prop: TAtom; var Value: Cardinal; Offset: Cardinal = 0): Boolean; overload;
function GetProperty(Dpy: PDisplay; Win: TWindow; Prop: TAtom; var Value: WideString; Offset: Cardinal = 0): Boolean; overload;
function GetProperty(Dpy: PDisplay; Win: TWindow; Prop: TAtom; var Value: string; Offset: Cardinal = 0): Boolean; overload;
function GetProperty(Dpy: PDisplay; Win: TWindow; Prop: TAtom; Value: TStrings; Offset: Cardinal = 0): Boolean; overload;

// Return the number of the virtual desktop where Win is visible.
function GetDesktop(Dpy: PDisplay; Win: TWindow): Cardinal;

// Return the window state.
function GetState(Dpy: PDisplay; Win: TWindow): Cardinal;

// Switch the current virtual desktop to Desktop.
function SetCurrentDesktop(Dpy: PDisplay; Desktop: Cardinal): Boolean;

const
  version = '$Id';
  
implementation

function GetProperty(Dpy: PDisplay; Win: TWindow; Prop: TAtom; var Value: Cardinal; Offset: Cardinal): Boolean; overload;
var
  ActualType: TAtom;
  ActualFormat: LongInt;
  NumItems, BytesAfter: Cardinal;
  Ptr: PByte;
begin
  Result := XGetWindowProperty(Dpy, Win, Prop, Offset, 1, LongBool(0), AnyPropertyType, @ActualType, @ActualFormat, @NumItems, @BytesAfter, @Ptr) = Success;
  if Result then
    try
      if (ActualType = None) or (ActualFormat <> 32) or not Assigned(Ptr) then
        Result := False;
      if Result then
        Value := PCardinal(Ptr)^;
    finally
      if Assigned(Ptr) then
        XFree(Ptr);
    end;
end;

function GetProperty(Dpy: PDisplay; Win: TWindow; Prop: TAtom; var Value: WideString; Offset: Cardinal): Boolean; overload;
var
  ActualType: TAtom;
  ActualFormat: LongInt;
  Length, NumItems, BytesAfter: Cardinal;
  Ptr: PByte;
  utf8value: string;
begin
  // Determine the length of the property value.
  Result := XGetWindowProperty(Dpy, Win, Prop, Offset, 0, LongBool(0), None, @ActualType, @ActualFormat, @NumItems, @BytesAfter, @Ptr) = Success;
  if Result and Assigned(Ptr) then
    XFree(Ptr);
  if (ActualType = None) or (ActualFormat <> 8) then
    Result := False;

  if Result then
  try
    // Get the entire property value.
    Length := (BytesAfter + 3) div 4;
    Result := XGetWindowProperty(Dpy, Win, Prop, Offset, Length, LongBool(0), ActualType, @ActualType, @ActualFormat, @NumItems, @BytesAfter, @Ptr) = Success;
    if (ActualType = None) or (ActualFormat <> 8) or not Assigned(Ptr) then
      Result := False;
    if Result then
    begin
      SetString(utf8value, PChar(Ptr), NumItems);
      Value := UTF8Decode(utf8value);
    end;
  finally
    if Assigned(Ptr) then
      XFree(Ptr);
  end;
end;

function GetProperty(Dpy: PDisplay; Win: TWindow; Prop: TAtom; var Value: string; Offset: Cardinal = 0): Boolean; overload;
var
  ActualType: TAtom;
  ActualFormat: LongInt;
  Length, NumItems, BytesAfter: Cardinal;
  Ptr: PByte;
begin
  // Determine the length of the property value.
  Result := XGetWindowProperty(Dpy, Win, Prop, Offset, 0, LongBool(0), None, @ActualType, @ActualFormat, @NumItems, @BytesAfter, @Ptr) = Success;
  if Result and Assigned(Ptr) then
    XFree(Ptr);
  if (ActualType = None) or (ActualFormat <> 8) then
    Result := False;

  if Result then
  try
    // Get the entire property value.
    Length := (BytesAfter + 3) div 4;
    Result := XGetWindowProperty(Dpy, Win, Prop, Offset, Length, LongBool(0), ActualType, @ActualType, @ActualFormat, @NumItems, @BytesAfter, @Ptr) = Success;
    if (ActualType = None) or (ActualFormat <> 8) or not Assigned(Ptr) then
      Result := False;
    if Result then
      SetString(Value, PChar(Ptr), NumItems);
  finally
    if Assigned(Ptr) then
      XFree(Ptr);
  end;
end;

procedure SetStringList(Strings: TStrings; Ptr: PChar; Count: Cardinal);
var
  Str: string;
begin
  while Count > 0 do
  begin
    Str := Ptr;
    Strings.Add(Str);
    Inc(Ptr,   Length(Str) + 1);
    Dec(Count, Length(Str) + 1);
  end;
end;

function GetProperty(Dpy: PDisplay; Win: TWindow; Prop: TAtom; Value: TStrings; Offset: Cardinal): Boolean; overload;
var
  ActualType: TAtom;
  ActualFormat: LongInt;
  Length, NumItems, BytesAfter: Cardinal;
  Ptr: PByte;
begin
  // Determine the length of the property value.
  Result := XGetWindowProperty(Dpy, Win, Prop, Offset, 0, LongBool(0), None, @ActualType, @ActualFormat, @NumItems, @BytesAfter, @Ptr) = Success;
  if Result and Assigned(Ptr) then
    XFree(Ptr);
  if (ActualType = None) or (ActualFormat <> 8) then
    Result := False;

  if Result then
  begin
    Value.BeginUpdate;
    try
      // Get the entire property value.
      Length := (BytesAfter + 3) div 4;
      Result := XGetWindowProperty(Dpy, Win, Prop, Offset, Length, LongBool(0), ActualType, @ActualType, @ActualFormat, @NumItems, @BytesAfter, @Ptr) = Success;
      if (ActualType = None) or (ActualFormat <> 8) or not Assigned(Ptr) then
        Result := False;
      if Result then
      begin
        Value.Clear;
        SetStringList(Value, PChar(Ptr), NumItems);
      end;
    finally
      Value.EndUpdate;
      if Assigned(Ptr) then
        XFree(Ptr);
    end;
  end;
end;

// Return the number of the virtual desktop where Win is visible.
// Return $FFFFFFFF if there are no virtual desktops. This is the value
// that represents a sticky desktop, too.
function GetDesktop(Dpy: PDisplay; Win: TWindow): Cardinal;
var
  A: TAtom;
begin
  A := XInternAtom(Dpy, '_NET_WM_DESKTOP', LongBool(1));
  if A = None then
    Result := $FFFFFFFF
  else if not GetProperty(Dpy, Win, A, Result) then
    Result := $FFFFFFFF;
end;

// Return the window state.
function GetState(Dpy: PDisplay; Win: TWindow): Cardinal;
var
  A: TAtom;
begin
  A := XInternAtom(Dpy, 'WM_STATE', LongBool(1));
  if A = None then
    Result := 0
  else if not GetProperty(Dpy, Win, A, Result) then
    Result := 0;
end;

// Change a property value.
function ChangeProperty(Dpy: PDisplay; Win: TWindow; Prop: TAtom; Value: Cardinal; Mode: TPropMode): Boolean; overload;
begin
  Result := XChangeProperty(Dpy, Win, Prop, XA_Cardinal, 32, Ord(Mode), PByte(@Value), 1) = Success;
end;

// Switch the current virtual desktop to Desktop.
function SetCurrentDesktop(Dpy: PDisplay; Desktop: Cardinal): Boolean;
var
  Net_Current_Desktop: TAtom;
  Root: TWindow;
  CurrentDesktop: Cardinal;
  Event: TXEvent;
begin
  Net_Current_Desktop := XInternAtom(Dpy, '_NET_CURRENT_DESKTOP', LongBool(1));
  if Net_Current_Desktop = None then
    Result := False
  else
  begin
    Root := XDefaultRootWindow(Dpy);
    // Make sure the root window has a desktop property.
    if not GetProperty(Dpy, Root, Net_Current_Desktop, CurrentDesktop) then
      Result := False
    else
    begin
      Event.xclient._type:=ClientMessage;
      Event.xclient.send_event := LongBool(1);
      Event.xclient.window := Root;
      Event.xclient.display := Dpy;
      Event.xclient.format := 32;
      Event.xclient.message_type := Net_Current_Desktop;
      Event.xclient.data.l[0] := Desktop;
      Result := XSendEvent(Dpy, Root, LongBool(0), SubstructureNotifyMask or SubstructureRedirectMask, @Event) = Success;
    end;
  end;
end;

// Enumerate all top-level windows.
// That is, call Proc for every window that is a top-level window.
// Pass Param untouched to Proc.
procedure EnumTopLevelWindows(Proc: TWindowProc; Param: Pointer);
  procedure QueryChildren(Dpy: PDisplay; Win: TWindow);
  var
    Root, Parent: TWindow;
    Children: PWindowArray;
    I, Count: Cardinal;
  begin
    // A top-level window might have been reparented to anywhere in the
    // window hierarchy. Thus, QueryChildren must recurse through all
    // window trees to find top-level windows.

    // Get all the children of Win.
    if XQueryTree(Dpy, Win, @Root, @Parent, PPWindow(@Children), @Count) = 0 then
      Exit;

    // Note that Count is a Cardinal, not an Integer, so the for loop
    // below cannot iterate through Count-1 (-1 not being a valid Cardinal),
    // so check explicitly for a zero Count.
    if Count = 0 then
      Exit;

    try
      // Loop through each child window.
      for I := 0 to Count-1 do
      begin
        QueryChildren(Dpy, Children[I]);
        // A top level window, by definition, has the WM_STATE property,
        // and the property value is non-zero.
        if GetState(Dpy, Children[I]) <> 0 then
          Proc(Dpy, Children[I], Param);
      end;
    finally
      XFree(Children);
    end;
  end;

var
  Display: PDisplay;
  Root: TWindow;

begin
  Display := nil;
  try

    // Start by opening the X-Windows display, if possible.
    Display := XOpenDisplay(Pchar(GetEnvironmentVariable('DISPLAY')));
    if not Assigned(Display) then
      Exit; // No X-Windows display.

    // Get the root window for the default screen.
    Root := XRootWindow(Display, XDefaultScreen(Display));
    // Get all the children of the root window.
    QueryChildren(Display, Root);
  finally
    if Assigned(Display) then
      XCloseDisplay(Display);
  end;
end;

end.
 
